// Author 		: F. de la Parra
// Date			: March, 2019
// Description 	: The HPOexplorer class implements the main flow of the 
//				  procedure to read in the HPO ontology and the queries 
//				  to run on it. It stores the navigation paths generated
//                by these queries in file "results.txt". Additionally,
//				  it implements the search for the largest path from a 
//				  "Term" node to the "root Term" node of the ontology. It
//				  stores this result in file "maxpath.txt". 
// Version		: 1.0

import java.util.Scanner;
import java.io.FileInputStream;
import java.io.PrintWriter;
import java.io.FileOutputStream;
import java.io.FileNotFoundException;

public class HPOExplorer {
	// Declare array of "Term" to hold HPO ontology in memory.
	Term[] terms = null;
	
	// Method "main" drives topic selection
	public static void main(String[] args) {
		
		// Create a 'demo' object.
		HPOExplorer HPOExplorerObject = new HPOExplorer();
		
		// Get number of terms in the HPO ontology;
		int termArraySize = HPOExplorerObject.getTermArraySize();
		
		// Initialize the size of "terms" array.
		HPOExplorerObject.terms = new Term[termArraySize];
		
		// Read the ontology into "terms" array.
		HPOExplorerObject.readHPO();
		
		// Create file "results.txt" with responses to the queries in "queries.txt".
		HPOExplorerObject.answerQueries();
		
		// Create file "maxpath.txt" with terms in a path to the root term of maximum length.
		HPOExplorerObject.findMaxPath();
	}
	
	// Accessor method for terms[] array.
	public Term[] getTerms() {
		return terms;
	}
	
	// Method to get the number of not obsolete terms in the HPO.txt ontology file.
	private int getTermArraySize() {
		Scanner HPO = null;
		int termCount = 0;
		try {
			String fileLine;
			boolean notObsolete = false;
			HPO = new Scanner(new FileInputStream("HPO.txt"));
			while (HPO.hasNextLine()) {
				fileLine = HPO.nextLine();
				if (fileLine.equals("[Term]"))
					notObsolete = true;
				if (fileLine.contains("is_obsolete"))
					notObsolete = false;
				if (fileLine.equals("") && notObsolete)
					termCount += 1;
			}
		}
		catch(FileNotFoundException e) {
			System.out.println("HPO.txt file not found. Program aborted!");
			System.exit(0);
		}
		HPO.close();
		return termCount;
	}
	
	// Method to read into array "terms" all not obsolete terms in the HPO.txt ontology file.
	private void readHPO() {
		int termCount = 0;
		Scanner HPO = null;
		try {
			int index = 0;
			String fileLine;
			StringBuilder term = new StringBuilder("[Term]\n");
			boolean notObsolete = false;
			boolean is_aFound = false;
			boolean blankLine = false;
			int first = 0;
			String id = "", is_a = "";
			HPO = new Scanner(new FileInputStream("HPO.txt"));
			
			// Skip the beginning of file HPO.txt
			fileLine = HPO.nextLine();
			while (!fileLine.equals("[Term]"))
				fileLine = HPO.nextLine();
			
			// Read the root term.
			while (!fileLine.equals("")) {
				fileLine = HPO.nextLine();
				if (fileLine.contains("id: HP:") && !fileLine.contains("alt_id:")) {
					first = fileLine.indexOf("HP:") + 3;
					id = fileLine.substring(first, first + 7);
				}	
				term.append(fileLine + "\n");
			}
			terms[index] = new Term(Integer.parseInt(id),
									term.substring(0, term.length() - 1));
			// System.out.print(term.substring(0));
			index += 1;
			
			// Read the remaining of the file.
			while (HPO.hasNextLine()) {
				fileLine = HPO.nextLine();
				if (fileLine.equals("[Term]")) {
					notObsolete = true;
					is_aFound = false;
					term.delete(0, term.length());
				}
				if (fileLine.contains("is_obsolete"))
					notObsolete = false;
				
				if (fileLine.contains("id: HP:") && !fileLine.contains("alt_id:")) {
					first = fileLine.indexOf("HP:") + 3;
					id = fileLine.substring(first, first + 7);
				}
				
				if (fileLine.contains("is_a: HP:") && !is_aFound) {
					first = fileLine.indexOf("HP:") + 3;
					is_a = fileLine.substring(first, first + 7);
					is_aFound = true;
				}	
				
				if (fileLine.equals("") && notObsolete && is_aFound) {
					terms[index] = new Term(Integer.parseInt(id),
											Integer.parseInt(is_a),
											term.substring(0));
					index += 1;
					term.delete(0, term.length() - 1);
				}
				else
					term.append(fileLine + "\n");
			}	
		}
		catch(FileNotFoundException e) {
			System.out.println("HPO.txt file not found. Program aborted!");
			System.exit(0);
		}
		HPO.close();
	}
	
	// This method reads the queries provided in "queries.txt" and store the 
	// navigation paths generated by these queries in "results.txt".
	private void answerQueries() {
		Scanner queries = null;
		try {
			queries = new Scanner(new FileInputStream("queries.txt"));
		}
		catch(FileNotFoundException e) {
			System.out.println("queries.txt file not found. Program aborted!");
			System.exit(0);
		}

		PrintWriter results = null;
		try {
			results = new PrintWriter(new FileOutputStream("results.txt"));
		}
		catch(FileNotFoundException e) {
			System.out.println("Cannot open file results.txt. Program aborted!");
			System.exit(0);
		}
		
		Query[] queryArray = new Query[5];		
		for (int i = 0; i < 5; i++) {
		queryArray[i] = new Query(queries.nextLine());
		results.write("[query_answer]\n");
		results.write(queryArray[i].getTermPath(getTerms()) + "\n");
		}
		
		queries.close();
		results.close();		
	}
	
	// This method finds the path to the root term, from any term in "HPO.txt",
	// which has the maximum length. It stores this path in "maxpath.txt".
	private void findMaxPath() {
		// Open for write file "maxpath.txt".
		PrintWriter maxpath = null;
		try {
			maxpath = new PrintWriter(new FileOutputStream("maxpath.txt"));
		}
		catch(FileNotFoundException e) {
			System.out.println("Cannot open file maxpath.txt. Program aborted!");
			System.exit(0);
		}
		
		// Find the term in the ontology whose path to the root is the longest.
		int maxPathLength = 0, currentPathLength;
		Term maxPathTerm = null;
		Query pathQuery = null;
		
		for (int i = 0; i < terms.length; i++) {
			pathQuery = new Query(terms[i].getId());
			currentPathLength = pathQuery.getPathLength(terms);
			if (currentPathLength > maxPathLength) {
				maxPathLength = currentPathLength;
				maxPathTerm = terms[i];
			}
		}
		
		// Store the terms on the maximum path in file "maxpath.txt".
		pathQuery = new Query(maxPathTerm.getId());
		maxpath.write("[max_path=" + maxPathLength + "]\n");
		maxpath.write(pathQuery.getTermPath(terms));
		maxpath.close();
	}
}